#include <cmath>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <iostream>
#include <sstream>
#include "matrix.hpp"
#include "rgbcolor.hpp"
#include "hsvcolor.hpp"

typedef float Real;
typedef SquareMatrix<Real> Matrix;

#define BLOCK_SIZE 16
#define GRID_SIZE 32
#define N (BLOCK_SIZE * GRID_SIZE)

__global__ void initialize(Matrix init) {
   unsigned int i = threadIdx.y + blockIdx.y * blockDim.y;
   unsigned int j = threadIdx.x + blockIdx.x * blockDim.x;

   const Real E_POWER_MINUS_PI = pow(M_E, -M_PI);
   Real value;
   if (i == N-1) {
      value = sin(M_PI * ((Real)j/N));
   } else if (i == 0) {
      value = sin(M_PI * ((Real)j/N)) * E_POWER_MINUS_PI;
   } else {
      value = 0;
   }
   init(i, j) = value;
}

__global__ void jacobi_iteration(Matrix source, Matrix target) {
   unsigned int i = threadIdx.y + blockIdx.y * blockDim.y;
   unsigned int j = threadIdx.x + blockIdx.x * blockDim.x;
   Real value;
   if (i == 0 || j == 0 ||
	 i + 1 >= blockDim.y * gridDim.y ||
	 j + 1 >= blockDim.x * gridDim.x) {
      value = source(i, j);
   } else {
      value = 0.25 *
	 (source(i-1, j) + source(i, j-1) + source(i, j+1) + source(i+1, j));
   }
   target(i, j) = value;
}

__global__ void color_converter(Matrix source, guchar* pixels,
      unsigned int rowstride, unsigned int n_channels) {
   unsigned int i = threadIdx.y + blockIdx.y * blockDim.y;
   unsigned int j = threadIdx.x + blockIdx.x * blockDim.x;
   Real value = source(i, j);
   float hue = (1 - value) * 240;
   HSVColor hsv_color(hue, 1, 1);
   RGBColor rgb_color = hsv_color.get_rgb();

   guchar* pp = pixels + i * rowstride + j * n_channels;
   pp[0] = rgb_color.red * 255;
   pp[1] = rgb_color.green * 255;
   pp[2] = rgb_color.blue * 255;
}

int main(int argc, char** argv) {
   const char* cmdname = *argv++; --argc;
   unsigned int max_iterations = 1000;
   if (argc > 1) {
      std::istringstream arg(*argv++); --argc;
      if (!(arg >> max_iterations)) {
	 std::cerr << cmdname << ": bad value for iterations" << std::endl;
	 exit(1);
      }
   }
   if (argc != 1) {
      std::cerr << "Usage: " << cmdname << " [iterations] out.png" << std::endl;
      exit(1);
   }

   const char* outfile = *argv++; --argc;

   Matrix m1; Matrix m2;
   m1.allocate_cuda_data(N);
   m2.allocate_cuda_data(N);

   // initialize matrix and run Jacobi iterations
   dim3 block_dim(BLOCK_SIZE, BLOCK_SIZE);
   dim3 grid_dim(GRID_SIZE, GRID_SIZE);
   initialize<<<grid_dim, block_dim>>>(m1);
   for (unsigned int iterations = 0; iterations < max_iterations;
	 iterations += 2) {
      jacobi_iteration<<<grid_dim, block_dim>>>(m1, m2);
      jacobi_iteration<<<grid_dim, block_dim>>>(m2, m1);
   }

   // visualize last iteration
   g_type_init();
   GdkPixbuf* pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB, FALSE, 8, N, N);
   int n_channels = gdk_pixbuf_get_n_channels(pixbuf);
   int rowstride = gdk_pixbuf_get_rowstride(pixbuf);
   guchar* pixels = gdk_pixbuf_get_pixels(pixbuf);
   if (!pixels) {
      std::cerr << cmdname << ": failed to allocate pixbuf with " <<
	 N << " x " << N << " pixels" << std::endl;
      exit(1);
   }
   guchar* cuda_pixels;
   size_t pixels_size = N * rowstride * sizeof(guchar);
   CHECK_CUDA(cudaMalloc, (void**)&cuda_pixels, pixels_size);
   color_converter<<<grid_dim, block_dim>>>(m1,
      cuda_pixels, rowstride, n_channels);
   CHECK_CUDA(cudaMemcpy, pixels, cuda_pixels, pixels_size,
      cudaMemcpyDeviceToHost);
   CHECK_CUDA(cudaFree, cuda_pixels);
   gdk_pixbuf_save(pixbuf, outfile, "png", (GError **)0, (void*)0);
}
