#ifndef MATRIX_H
#define MATRIX_H

#include <cassert>
#include <iostream>
#include "cuda.hpp"

template<typename T>
struct SquareMatrix {
   unsigned int N;
   T* data;
   T* cuda_data;
   bool cloned;

   SquareMatrix() : N(0), data(0), cuda_data(0), cloned(false) {
   }
   SquareMatrix(const SquareMatrix& other) : N(other.N), data(other.data),
	 cuda_data(other.cuda_data), cloned(true) {
   }
   ~SquareMatrix() {
      if (!cloned) {
	 if (data) delete data;
	 if (cuda_data) release_cuda_data();
      }
   }

   void copy_to_gpu() {
      if (!cuda_data) {
	 allocate_cuda_data();
      }
      CHECK_CUDA(cudaMemcpy, cuda_data, data, N * N * sizeof(T),
	 cudaMemcpyHostToDevice);
   }

   void copy_from_gpu() {
      if (!data) {
	 data = new T[N * N];
      }
      assert(cuda_data);
      CHECK_CUDA(cudaMemcpy, data, cuda_data, N * N * sizeof(T),
	 cudaMemcpyDeviceToHost);
   }

   void allocate_cuda_data() {
      if (!cuda_data) {
	 T* cudap;
	 CHECK_CUDA(cudaMalloc, (void**)&cudap, N * N * sizeof(T));
	 cuda_data = cudap;
      }
   }

   void allocate_cuda_data(unsigned int N_) {
      if (N != N_) {
	 if (data) {
	    delete data; data = 0;
	 }
	 release_cuda_data();
	 N = N_;
      }
      allocate_cuda_data();
   }

   void release_cuda_data() {
      if (cuda_data) {
	 CHECK_CUDA(cudaFree, cuda_data);
	 cuda_data = 0;
      }
   }

   void resize(unsigned int N_) {
      if (N != N_) {
	 T* rp = new T[N_ * N_];
	 if (data) delete data;
	 release_cuda_data();
	 data = rp; N = N_;
      }
   }

   HOST_DEV T& operator()(unsigned int i, unsigned int j) {
#ifdef __CUDA_ARCH__
      return cuda_data[i*N + j];
#else
      return data[i*N + j];
#endif
   }

   HOST_DEV const T& operator()(unsigned int i, unsigned int j) const {
#ifdef __CUDA_ARCH__
      return cuda_data[i*N + j];
#else
      return data[i*N + j];
#endif
   }
};

// input operator for square matrices that expects
// an input sequence consisting of N, followed by N*N
// elements of the matrix in row-major order
template<typename T>
static inline std::istream& operator>>(std::istream& in, SquareMatrix<T>& matrix) {
   unsigned int N;
   if (!(in >> N)) return in;
   if (N == 0) {
      in.setstate(std::ios::failbit); return in;
   }
   matrix.resize(N);
   for (int i = 0; i < N; ++i) {
      for (int j = 0; j < N; ++j) {
	 if (!(in >> matrix(i, j))) return in;
      }
   }
   return in;
}

template<typename T>
static inline std::ostream& operator<<(std::ostream& out, SquareMatrix<T>& matrix) {
   unsigned int N = matrix.N;
   for (int i = 0; i < N; ++i) {
      for (int j = 0; j < N; ++j) {
	 if (!(out << " " << matrix(i, j))) return out;
      }
      out << std::endl;
   }
   return out;
}

#endif
